use std::net::SocketAddr;

use bus::Bus;

mod util;
mod app;
mod system;
mod protocol;

use app::App;
use system::{cache::WebCache, net::Stream};
use tokio_stream::StreamExt;
use tokio_tungstenite::WebSocketStream;
use hyper_util::rt::TokioIo;
use tokio_rustls::server::TlsStream;
use tokio::net::TcpStream;

async fn thread_datasystem(mut _app:App, bus:Bus<protocol::QRPacket>)
{
    use protocol::QRPacket;
    use game::protocol::packet::*;

    while match bus.receive_wait() {
        Some(packet) => {
            match packet.data {
                QRPacket::QAuth(_request) => {
                    let response = PacketAuthResponse::new();

                    // get user id from handle

                    // get user salt

                    // hash request secret

                    // compare hash to user secret
                    
                    // return response
                    bus.send(packet.from, QRPacket::RAuth(response)).is_ok()
                }
                _ => { true }
            }
        }
        None => false,
    } { }
}

#[derive(Clone)]
struct HttpServiceArgs {
    bus:Bus<protocol::QRPacket>,
    cache:WebCache,
}

async fn handle_ws(mut ws_stream:WebSocketStream<TlsStream<TcpStream>>, args:HttpServiceArgs)
{
    use tokio_tungstenite::tungstenite::protocol::Message;
    use game::util::unpack_u16;
    use protocol::QRPacket;
    use game::protocol::{
        code::*,
        packet::*,
    };

    let bus_ds = args.bus.mailbox(1).unwrap_or(1);

    while match ws_stream.try_next().await {
        Ok(msg) => match msg {
            Some(Message::Binary(data)) => {
                
                let mut index :usize = 0;
                let code: u16 = unpack_u16(&data, &mut index);

                match code {
                    CODE_AUTH => {
                        match PacketAuth::decode(&data, &mut index) {
                            Ok(packet) => {
                                if args.bus.send(bus_ds, QRPacket::QAuth(packet)).is_ok() {
                                    match args.bus.receive_wait() {
                                        Some(resp) => match resp.data {
                                            QRPacket::RAuth(_resp) => {
                                                
                                            }
                                            _ => { }
                                        }
                                        None => { }
                                    }
                                }
                            }
                            Err(_) => { }
                        }
                    }
                    _ => { }
                }

                true
            }
            Some(_) => true,
            None => false
        }
        Err(_) => false,
    } { }

    ws_stream.close(None).await.ok();
}

async fn service_http(request:hyper::Request<hyper::body::Incoming>, args:HttpServiceArgs) -> Result<hyper::Response<String>, std::convert::Infallible>
{
    use hyper::Response;
    use tokio_tungstenite::accept_async;

    match request.uri().path() {
        "" => Ok(Response::new(args.cache.html())),
        ".css" => Ok(Response::new(args.cache.css())),
        ".js" => Ok(Response::new(args.cache.js())),
        ".wasm" => Ok(Response::new(args.cache.wasm())),
        //"favicon.png" => Ok(Response::new(String::new())),
        "ws" => {
            if request.headers().get(hyper::header::UPGRADE).map(|h| h == "websocket").unwrap_or(false) {
                match hyper::upgrade::on(request).await {
                    Ok(upgraded) => {
                        match upgraded.downcast::<TokioIo<TlsStream<TcpStream>>>() {
                            Ok(parts) => {
                                match accept_async(parts.io.into_inner()).await {
                                    Ok(ws_stream) => { handle_ws(ws_stream, args).await }
                                    Err(_) => { }
                                }
                            }
                            Err(_) => { }
                        }
                    }
                    Err(_) => { }
                }
            }
            Ok(Response::builder()
                .status(101)
                .header(hyper::header::UPGRADE, "websocket")
                .body(String::new())
                .unwrap())
        }
        _ => Ok(Response::new(String::new())),
    }
}

async fn handle_http(stream:system::net::tls::TlsStream, addr:SocketAddr, args:HttpServiceArgs) -> Result<(),()>
{
    use hyper::server::conn::http1;
    use hyper::service::service_fn;

    println!("Connection from {}", addr.to_string());

    let io = TokioIo::new(stream.to_stream());

    http1::Builder::new()
        .serve_connection(io, service_fn(move |req| {
            service_http(req, args.clone())
        }))
        .await.ok();
    Ok(())
}


#[tokio::main]
async fn main()
{
    use system::net::{
        Server,
        tls::*,
    };

    // Initialize application data.
    let mut app = App::new();
    if app.init().is_err() {
        println!("fatal: failed to initialize server.");
        return;
    }


    // Initialize central bus and data serivce.
    let b_main :Bus<protocol::QRPacket> = Bus::new_as(bus::Mode::Transmitter);
    match b_main.connect() {
        Ok(bus) => {
            tokio::spawn(async move {
                thread_datasystem(app, bus).await;
            });
        }
        Err(_) => {
            println!("fatal: failed to initialize bus.");
            return;
        }
    }


    // Initialize HTTPS service.
    match b_main.connect() {
        Ok(bus) => {
            let cache = WebCache::init();

            let mut server = TlsServer::new();
            server.add_cert("omen.kirisame.com", "cert/fullchain.pem", "cert/privkey.pem").await.ok();
            match server.bind("0.0.0.0:38612").await {
                Ok(_) => {
                    tokio::spawn(async move {
                        while server.accept(handle_http, HttpServiceArgs {
                            bus:bus.connect().unwrap(),
                            cache:cache.clone(),
                        }).await.is_ok() { }
                    });
                }
                Err(_) => {
                    println!("error: failed to bind port 38612.");
                }
            }
        }
        Err(_) => {
            println!("error: failed to initialize HTTPS service.");
        }
    }

    std::thread::park();
}
